/**
 * @file
 * Chat controller.
 */

(function () {
  'use strict';

  angular.module('PeerJSChat')
    .controller('ChatCtrl', function ($scope, $sce, $interval, $timeout, $filter, $http, EntityService, entity, user, options, constraints, baseUrl) {
      $scope.pid = user.pid;
      $scope.Peer = null;
      $scope.localStream = null;
      $scope.peers = [];
      $scope.mediaConnections = [];
      $scope.dataConnections = [];
      $scope.messages = [];
      $scope.alerts = [];
      $scope.formData = {};
      $scope.formData.status = localStorage.getItem('peerStatus') || 1;
      $scope.formData.videoChatStatus = 0;
      var peersInterval;

      /**
       * Initialize controller.
       */
      $scope.init = function () {
        $scope.user = user;
        if ($scope.formData.status == 1) {
          $scope.changeStatus(1);
        }
      };

      /**
       * Initialize controller.
       */
      $scope.changeStatus = function (status) {
        $scope.formData.status = status;
        localStorage.setItem('peerStatus', status);
        if (status) {
          $scope.createPeerConnection();
        } else {
          $scope.destroyPeerConnection();
        }
      };

      /**
       * Create the Peer connection.
       */
      $scope.createPeerConnection = function () {
        $scope.Peer = new Peer(null, options);

        // On open connection to the PeerServer.
        $scope.Peer.on('open', function () {

          // Add user to this chat room. Creates a new peer in the
          // peerjs_peer table.
          EntityService.get({
            type: 'peerjs_peer',
            cid: entity.cid,
            author: user.uid
          })
            .$promise.then(function (data) {

              // Create a new peer entry if it doesn't exist.
              if (data.list.length === 0) {
                new EntityService({
                  pid: $scope.Peer.id,
                  author: user.uid,
                  cid: entity.cid
                }).$save({type:'peerjs_peer'});
              }

              // Otherwise update the old entry.
              else if (data.list.length === 1) {
                EntityService.update({
                  type: 'peerjs_peer',
                  id: data.list[0].id
                }, {
                  pid: $scope.Peer.id,
                })
                  .$promise.then(null, function (error) {
                    console.log(error);
                    console.log('Attempting to update peer using GET request...');

                    // Temporary work around for servers that don't allow PUT/DELETE requests.
                    $http.post(baseUrl + '/peerjs/update', {
                      id: data.list[0].id,
                      pid: $scope.Peer.id
                    });
                  });
              }
            }, function (error) {
              console.log(error);
            });

          // Attempt to establish a data connection with each peer.
          $scope.connectPeers();
        });

        // On close connection to the PeerServer.
        $scope.Peer.on('close', function () {
          $scope.mediaConnections = [];
          $scope.dataConnections = [];
          $scope.calls = [];
          $scope.messages = [];
        });

        // On disconnect from the PeerServer.
        $scope.Peer.on('disconnected', function (peer) {
        });

        // Handle incoming calls.
        $scope.Peer.on('call', function (mediaConnection) {
          console.log("CALL RECEIVED:", mediaConnection);
          if (mediaConnection.metadata.type === 'public') {
            $scope.publicAnswer(mediaConnection);
          }
          if (mediaConnection.metadata.type === 'private') {
            $scope.privateAnswer(mediaConnection);
          }
        });

        // Handle incoming data connections.
        $scope.Peer.on('connection', function (dataConnection) {

          // Only connect to new peers.
          if ($filter('filter')($scope.dataConnections, {'peer': dataConnection.peer}).length === 0) {
            $scope.initializeDataConnection(dataConnection);
          } else {
            dataConnection.close();
          }
        });

        // Handle errors.
        $scope.Peer.on('error', function (error) {
          if (error.type === 'browser-incompatible') {
            $scope.changeStatus(0);
            $scope.alerts = [];
            $scope.addAlert('danger', error.message + '. Please try Google Chrome or Firefox instead.', 0);
          }
        });

        // Send a heartbeat to Peer Server to keep the connection open.
        peersInterval = $interval(function () {
          if ($scope.Peer.socket._wsOpen()) {
            $scope.Peer.socket.send({type: 'HEARTBEAT'});
            $scope.connectPeers();
            if ($scope.formData.videoChatStatus) {
              $scope.publicCall();
            }
          }
        }, 5000);
      };

      $scope.addAlert = function(type, message, duration) {
        var alert = {
          message: message,
          type: type
        };
        $scope.alerts.push(alert);

        if (duration) {
          $timeout(function () {
            $scope.closeAlert(alert);
          }, duration);
        }
      };

      $scope.closeAlert = function(alert) {
        $scope.alerts.splice($scope.alerts.indexOf(alert), 1);
      };

      /**
       * Close the Peer connection.
       */
      $scope.destroyPeerConnection = function () {
        if ($scope.Peer !== null) {
          $scope.Peer.destroy();
          $scope.Peer = null;
          $scope.peers = [];
          $scope.mediaConnections = [];
          $scope.dataConnections = [];
          $scope.messages = [];
          $interval.cancel(peersInterval);
          $scope.stopLocalMediaStream();
          $scope.formData.videoChatStatus = false;
        }
      };

      /**
       * Initialize controller.
       */
      $scope.connectPeers = function () {
        EntityService.get({
          type: 'peerjs_peer',
          cid: entity.cid,
        })
          .$promise.then(function (data) {
            $scope.peers = $filter('filter')(data.list, {'uid': '!' + $scope.user.uid});

            // Iterate over each peer and try to connect to peer.
            data.list.forEach(function (peer) {

              // Only try to connect to new peers.
              if ($filter('filter')($scope.dataConnections, {'peer': peer.pid}).length === 0) {
                $scope.initializeDataConnection($scope.Peer.connect(peer.pid));
              }
            });
          });
      };

      /**
       * Create a dataConnection with a peer.
       */
      $scope.initializeDataConnection = function (dataConnection) {

        // Handle dataConnection open event.
        dataConnection.on('open', function () {

          // Only add this connection if it doesn't already exist.
          if ($filter('filter')($scope.dataConnections, {'peer': this.peer}).length === 0) {

            // Let peer know that connection was successful, and send user info.
            this.send({
              type: 'status',
              user: user
            });
          }
        });

        // Handle dataConnection close event.
        dataConnection.on('close', function () {
          var index = $scope.dataConnections.indexOf(this);
          $scope.dataConnections.splice(index, 1);
        });

        // Handle dataConnection data event.
        dataConnection.on('data', function (data) {
          if (data.type === 'status') {
            this.metadata = {
              user: data.user
            };
            $scope.dataConnections.push(this);
          }
          if (data.type === 'public') {
            $scope.messages.push(data);
          }
        });
      };

      /**
       * Create a dataConnection with a peer.
       */
      $scope.message = function () {
        var message = {
          type: 'public',
          user: user,
          text: $scope.formData.messageText
        };

        $scope.messages.push(message);
        $scope.dataConnections.forEach(function (dataConnection) {
          message.peer = dataConnection.peer;
          dataConnection.send(message);
        });

        $scope.formData.messageText = '';
      };

      /**
       * Initialize controller.
       */
      $scope.updateStatus = function () {
        $scope.dataConnections.forEach(function (dataConnection) {
        });
      };

      /**
       * Enable the local media stream.
       */
      $scope.startLocalMediaStream = function (cb) {
        $scope.stopLocalMediaStream();
        navigator.getUserMedia(
          constraints,
          function(stream) {
            $scope.localStream = stream;
            $scope.localStreamUrl = $sce.trustAsResourceUrl(URL.createObjectURL(stream));
            if (angular.isFunction(cb)) {
              cb();
            }
          },
          function (error) {
            console.log(error);
          }
        );
      };

      /**
       * Disable the local media stream.
       */
      $scope.stopLocalMediaStream = function () {
        if ($scope.localStream !== null) {
          $scope.localStream.stop();
          $scope.localStream = null;
          $scope.localStreamUrl = '';
        }
      };

      /**
       * Call a peer.
       */
      $scope.joinVideoChat = function () {
        $scope.formData.videoChatStatus = confirm('Are you sure you want to join video chat? Your webcam video will be broadcast to all members in this chat room.');
        if ($scope.formData.videoChatStatus) {
          $scope.publicCall();
        }
      };

      /**
       * Call a peer.
       */
      $scope.leaveVideoChat = function () {
        $scope.formData.videoChatStatus = !confirm('Are you sure you want to leave video chat?');
        if (!$scope.formData.videoChatStatus) {
          $scope.publicHangup();
        }
      };

      /**
       * Call a peer.
       */
      $scope.publicCall = function () {
        if ($scope.localStream === null) {
          $scope.startLocalMediaStream(function () {
            $scope.publicCall();
          });
          return;
        }
        $scope.dataConnections.forEach(function (dataConnection) {
          var filteredMediaConnections = $filter('filter')($scope.mediaConnections, {'peer':dataConnection.peer, metadata: {user: {uid: user.uid}}});
          if (filteredMediaConnections.length === 0) {
            var mediaConnection = $scope.Peer.call(dataConnection.peer, $scope.localStream, {
              metadata: {
                type: 'public',
                user: user
              }
            });
            $scope.initializeMediaConnection(mediaConnection);
          }
        });
      };

      /**
       * Call a peer.
       */
      $scope.publicHangup = function () {
        var filteredMediaConnections = $filter('filter')($scope.mediaConnections, {metadata: {user: {uid: user.uid}}});
        console.log(filteredMediaConnections);
        filteredMediaConnections.forEach(function (mediaConnection) {
          mediaConnection.close();
        });
        $scope.stopLocalMediaStream();
      };

      /**
       * Answer a public call.
       */
      $scope.publicAnswer = function (mediaConnection) {
        var filteredMediaConnections = $filter('filter')($scope.mediaConnections, {'peer': mediaConnection.peer, answered: true, open: true});
        if (filteredMediaConnections.length === 0) {
          mediaConnection.answered = true;
          mediaConnection.answer(null);
          $scope.initializeMediaConnection(mediaConnection);
          console.log("ANSWERED:", mediaConnection);
        } else {
          console.log("NOT ANSWERED:", mediaConnection);
        }

      };

      /**
       * Call a peer.
       */
      $scope.privateCall = function (peer) {
        if ($scope.localStream === null) {
          $scope.startLocalMediaStream(function () {
            $scope.privateCall(peer);
          });
          return;
        }
        var mediaConnection = $scope.Peer.call(peer, $scope.localStream, {
          metadata: {
            type: 'private',
            user: user
          }
        });
        $scope.initializeMediaConnection(mediaConnection);
      };

      /**
       * Answer a peer call.
       */
      $scope.privateAnswer = function (mediaConnection) {
        if (!mediaConnection.metadata.accepted) {
          mediaConnection.metadata.accepted = confirm('You have a call from ' + mediaConnection.metadata.name + ', would you like to accept?');
        }
        if (mediaConnection.metadata.accepted) {
          if ($scope.localStream === null) {
            $scope.startLocalMediaStream(function () {
              $scope.privateAnswer(mediaConnection);
            });
            return;
          }
          mediaConnection.answer($scope.localStream);
          $scope.initializeMediaConnection(mediaConnection);
        }
      };

      /**
       * Create a dataConnection with a peer.
       */
      $scope.initializeMediaConnection = function (mediaConnection) {

        // Handle mediaConnection open event.
        // NOTE: This event doesn't seem to fire correctly.
        mediaConnection.on('open', function () {
          console.log("MEDIA CONNECTION OPEN");
        });

        // Handle mediaConnection close event.
        mediaConnection.on('close', function () {
          var index = $scope.mediaConnections.indexOf(mediaConnection);
          $scope.mediaConnections.splice(index, 1);
        });

        $scope.mediaConnections.push(mediaConnection);
      };
    });
}());


